﻿using System;
using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Vintagestory.API.Server;
using Vintagestory.API.Config;
using VSSurvivalMod;
using Vintagestory.GameContent;
using Vintagestory.API.Datastructures;

[assembly: ModInfo( "MortalDamage",
	Description = "More realistic damage system based around the game Earthbound",
	Website     = "",
	Authors     = new []{ "OmegaHaxors" } )]

namespace MortalDamage
{

	public class MortalDamage: ModSystem
	{
		int bleedtickrate = 500; //how long (in MS) it takes to tick the drawblood function. Does not impact balance, only performance. Bigger is better but slower.
		int bleedratemultiplier = 1; //how fast you bleed out. These both default to 1x. Change these value to affect the balance.
		int regenratemultiplier = 1; //how fast you regenerate compared to the original balance.
		ICoreAPI api = null;
		ICoreServerAPI sapi = null;
		ICoreClientAPI capi = null;
		

		public override void Start(ICoreAPI api){this.api = api;}
		public override void StartClientSide(ICoreClientAPI capi)
		{
			this.capi = capi;
			capi.Gui.RegisterDialog(new GuiDialog[]
			{new HudElementBloodBar(capi)});
		
		}
		public override void StartServerSide(ICoreServerAPI sapi)
		{
			this.sapi = sapi;
			sapi.Event.PlayerNowPlaying += OnPlayerJoin;
			sapi.Event.PlayerRespawn += OnPlayerRespawn;
			sapi.Event.PlayerDeath += OnPlayerDie;
			sapi.Event.RegisterGameTickListener(BleedTick,bleedtickrate);
			sapi.RegisterCommand("bleed","Tells you how much bleed and regen you have","",
			(IServerPlayer splayer, int groupId, CmdArgs args)
			=>{
			EntityBehaviorHealth healthbehavior = splayer.Entity.GetBehavior<EntityBehaviorHealth>();
			if(healthbehavior == null || splayer.Entity.Attributes == null){splayer.SendMessage(GlobalConstants.GeneralChatGroup,"!?!",EnumChatType.OwnMessage); return;}
			string message = "Current Health: ";
			float health = (float)Math.Round(healthbehavior.Health*100)/100;
			message = message + health + "/" + Math.Round(healthbehavior.MaxHealth*100)/100 + "<br>";

			float bleed = (float)Math.Round(splayer.Entity.WatchedAttributes.GetFloat("bleed")*100)/100;
			float regen = (float)Math.Round(splayer.Entity.WatchedAttributes.GetFloat("regen")*100)/100;
			float truehealth = health - bleed;
			if(truehealth > 0){truehealth = truehealth + regen;}
			message = message + "True Health: ";
			message = message + truehealth + "/" + Math.Round(healthbehavior.MaxHealth*100)/100;
			if(bleed != 0){
			message = message + "<br>" + "Bleed: ";
			message = message + bleed;}
			if(regen != 0){
			message = message + "<br>" + "Regen: ";
			message = message + regen;}

			splayer.SendMessage(GlobalConstants.GeneralChatGroup,message,EnumChatType.OwnMessage);
			
			});
		}
		
		private void OnPlayerJoin(IServerPlayer byPlayer){
			EntityBehaviorHealth bhealth = byPlayer.Entity.GetBehavior<EntityBehaviorHealth>();
			if (bhealth != null){bhealth.onDamaged += ((float dmg, DamageSource dmgSource) => this.handleDamaged(byPlayer, dmg, dmgSource));}
			if (byPlayer.Entity !=null && byPlayer.Entity.Attributes!= null && byPlayer.Entity.WatchedAttributes.GetString("mortallywoundedby") == null){byPlayer.Entity.WatchedAttributes.SetString("mortallywoundedby","");}
			}
		private void OnPlayerDie(IServerPlayer byPlayer,DamageSource damageSource)
		{
			if(damageSource.Type == EnumDamageType.Poison && damageSource.Source == EnumDamageSource.Void && byPlayer.Entity.Attributes != null && byPlayer.Entity.WatchedAttributes.GetString("mortallywoundedby") != "")
			{
				string message = byPlayer.PlayerName + " has bled to death because of " + byPlayer.Entity.WatchedAttributes.GetString("mortallywoundedby");
				sapi.SendMessageToGroup(GlobalConstants.GeneralChatGroup,message,EnumChatType.Notification);
			}
		}
		private void OnPlayerRespawn(IServerPlayer byPlayer)
		{
			byPlayer.Entity.WatchedAttributes.SetFloat("bleed",0f); //prevent damage/regen from rolling over to the new life
			byPlayer.Entity.WatchedAttributes.SetFloat("regen",0f);
		}

		private float handleDamaged(IServerPlayer player,float damage, DamageSource source){
			EnumDamageType type = source.Type;
			float tobleed = damage;

			switch(type)
			{
				case EnumDamageType.Heal:
				return handleHealed(player,damage,source);

				case EnumDamageType.BluntAttack:
				tobleed = damage * 0.9f; //50% of the damage is taken directly, the rest becomes bleed
				break;

				case EnumDamageType.PiercingAttack:
				tobleed = damage * 0.9f; //25% of the damage is taken directly
				break;

				case EnumDamageType.SlashingAttack:
				tobleed = damage * 0.9f; //10% of the damage taken directly
				break;

				case EnumDamageType.Gravity:
				tobleed = damage * 0.5f; //50% of damage taken from falls
				break;

				default:
				return damage;
			}
			//just going to revert it back to 90% bleed for all, except gravity of course
			//The game just isn't set up to handle different damage types properly at the moment
			if(source.Source == EnumDamageSource.Internal || source.Source == EnumDamageSource.Suicide){return damage;}
			
				player.Entity.WatchedAttributes.SetFloat("bleed",player.Entity.WatchedAttributes.GetFloat("bleed")+tobleed); //apply bleed
				damage = damage - tobleed; //then reduce damage by the amount converted into bleed
				if(player.Entity.HasBehavior("health")){
					if(player.Entity.GetBehavior<EntityBehaviorHealth>().Health <= damage + player.Entity.WatchedAttributes.GetFloat("bleed"))
					{//player has taken fatal damage and will bleed out if medical attention is not given
						sapi.World.PlaySoundFor(new AssetLocation("mortaldamage:sounds/mortalhit"),player,false,32f,1f);
						if(source.SourceEntity != null)
						{
						player.Entity.WatchedAttributes.SetString("mortallywoundedby",source.SourceEntity.GetName());
						}else{
						player.Entity.WatchedAttributes.SetString("mortallywoundedby","Fall Damage"); //just assume fall damage if it's null
						}
						player.SendIngameError("mortaldamage","You received mortal damage!");
					}else
					{//it's not a fatal hit, so tell the player how much HP they will have left instead
						player.SendIngameError("bleeddamage","You will bleed to "+(Math.Round(player.Entity.GetBehavior<EntityBehaviorHealth>().Health - player.Entity.WatchedAttributes.GetFloat("bleed"),2))+"HP");
					}
				}
				return damage;
		}
		private float handleHealed(IServerPlayer player,float toheal, DamageSource source){
			if(source.Source == EnumDamageSource.Revive) {return toheal;} //gain an instant full heal from respawning
			if(player.Entity.HasBehavior("health") == false){return toheal;} //??? Just to be safe, lets check it here.
			if(player.Entity.WatchedAttributes.HasAttribute("healingeffectivness"))
			{toheal *= player.Entity.WatchedAttributes.GetFloat("healingeffectiveness");} //account for armor modifier
			float bleed = player.Entity.WatchedAttributes.GetFloat("bleed");
			float regen = player.Entity.WatchedAttributes.GetFloat("regen");
			//if(bleed <= 0){player.SendIngameError("uselessbandage","Health gain can only be used to reduce bleeding."); return 0f;}
			if(bleed <= 0) //apply regen instead
			{regen += toheal; player.Entity.WatchedAttributes.SetFloat("regen",regen);
			player.SendIngameError("regenvalue","You will slowly regenerate to "+Math.Round(Math.Min(player.Entity.GetBehavior<EntityBehaviorHealth>().MaxHealth,player.Entity.GetBehavior<EntityBehaviorHealth>().Health + player.Entity.WatchedAttributes.GetFloat("regen")),2)+ "HP");
			return 0f;}
			bleed = bleed - toheal;
			if(player.Entity.GetBehavior<EntityBehaviorHealth>().Health > bleed
			&& player.Entity.Attributes != null && player.Entity.WatchedAttributes.GetString("mortallywoundedby") != "")
			{
				player.SendIngameError("lifesaver","That just saved your life.");
				player.Entity.WatchedAttributes.SetString("mortallywoundedby","");
			}
			if(bleed <= 0){regen = regen - bleed; bleed = 0; player.SendIngameError("bleedingstoppedregen",
			"The bleeding has stopped. You will slowly regenerate to "+ Math.Round(Math.Min(player.Entity.GetBehavior<EntityBehaviorHealth>().MaxHealth,player.Entity.GetBehavior<EntityBehaviorHealth>().Health + player.Entity.WatchedAttributes.GetFloat("regen")),2)+ "HP");}
			player.Entity.WatchedAttributes.SetFloat("bleed",bleed); //remove bleed
			player.Entity.WatchedAttributes.SetFloat("regen",regen); //add regen
			return 0f;
		}
		private void BleedTick(float par){
			IServerPlayer[] players = (IServerPlayer[])sapi.World.AllOnlinePlayers;
			foreach (IServerPlayer player in players)
			{
				EntityPlayer eplayer = player.Entity;
				if(eplayer == null || player.ConnectionState != EnumClientState.Playing){continue;} //because the game consideres players still logging in to be 'online'
				if(eplayer.Attributes != null && eplayer.Properties != null && eplayer.HasBehavior("health"))
				{
					float bleed = eplayer.WatchedAttributes.GetFloat("bleed");
					float regen = eplayer.WatchedAttributes.GetFloat("regen");
					EntityBehaviorHealth bhealth = eplayer.GetBehavior<EntityBehaviorHealth>();
					if(bleed == float.Epsilon || float.IsNaN(bleed) || float.IsInfinity(bleed)){bleed = 0f;}
					if(bleed == float.Epsilon || float.IsNaN(bleed) || float.IsInfinity(bleed)){regen = 0f;}
					
					
					if(bleed >= 0.001f) //is the player bleeding?
					{
						float bloodloss = (bleed * 0.001f) + 0.001f;
						bloodloss *= (bleedtickrate/10) * bleedratemultiplier;
						if(!eplayer.Controls.Sprint){bloodloss /= 2;}//running causes you to lose blood faster
						if(eplayer.Controls.Sneak){bloodloss /= 2;}//earthbound reference, defending lowers health loss
						DamageSource damagesource = new DamageSource();
						damagesource.Source = EnumDamageSource.Void; damagesource.Type = EnumDamageType.Poison; damagesource.KnockbackStrength = 0f;
						bhealth.Health = bhealth.Health - bloodloss; //take away health manually
						if(bhealth.Health <= 0){bhealth.Health = 0; eplayer.ReceiveDamage(damagesource,1);} //kill the player if they're below 0HP
						bleed = bleed - bloodloss; //remove bleed damage now that you've suffered from it
						if(bleed <= 0.001f)//round it down to zero then tell the player they're not bleeding.
						{
							bleed = 0;
							if(regen > 0)
							{
							player.SendIngameError("bleedingstoppedregen",
							"The bleeding has stopped. You will slowly regenerate to "+ Math.Round(Math.Min(bhealth.MaxHealth,bhealth.Health + regen),2)+ "HP");
							}else{
							player.SendIngameError("bleedingstopped","The bleeding has stopped.");
							}
						}
						eplayer.WatchedAttributes.SetFloat("bleed",bleed);
					}else if(regen >= 0.0001f){//regen code runs if the player is not bleeding
						float healthgain = (regen * 0.0001f) + 0.0001f;
						healthgain *= (bleedtickrate/10) * regenratemultiplier;
						bhealth.Health = bhealth.Health + healthgain; //apply regeneration manually
						regen = regen - healthgain;
						if(bhealth.Health >= bhealth.MaxHealth)
						{
							bhealth.Health = bhealth.MaxHealth; regen = 0f;
							player.SendIngameError("maxedout","Your HP was maxed out.");
						} //waste regen if player is max hp
						eplayer.WatchedAttributes.SetFloat("regen",regen);
					}
					bhealth.MarkDirty(); //send a packet to the client
					
				}
			}
		}
	}

	public class HudElementBloodBar : HudElement //credit to OxygenNotIncluded. I want to change this into an arrow that points, but GUI code suuuuucks.
	{
		double BloodBarVerticalAlignmentOffset = 0; // normally -5
		public override double InputOrder
		{
			get
			{
				return 1;
			}
		}
		public override double DrawOrder
		{
			get
			{
				return 0.05;
			}
		}

		public HudElementBloodBar(ICoreClientAPI capi) : base(capi)
		{
			capi.Event.RegisterGameTickListener(new Action<float>(this.OnGameTick), 20, 0);
		}

		private void OnGameTick(float dt)
		{
			this.UpdateBloodBar();
		}

		private void UpdateBloodBar()
		{
			if (this.statbar == null || capi.World.Player.Entity.WatchedAttributes.GetTreeAttribute("health") == null)
			{return;}
			SyncedTreeAttribute treeAttribute = this.capi.World.Player.Entity.WatchedAttributes;
			float? bleed = (treeAttribute != null) ? treeAttribute.TryGetFloat("bleed") : null;
			float? regen = (treeAttribute != null) ? treeAttribute.TryGetFloat("regen") : null;
			ITreeAttribute healthtree = treeAttribute.GetTreeAttribute("health");
			float health = healthtree.GetFloat("currenthealth");
			float max = healthtree.GetFloat("maxhealth",20);
			if (bleed == null){bleed = 0;} if(regen == null){regen = 0;}
			if(max == 0){return;}//don't divide by zero
			float lineInterval = (float)(max * (1/Math.Round(max - 0.5)));
			this.statbar.SetLineInterval(lineInterval);
			float barvalue = 0;
			if(bleed > 0){ //bleed code
			barvalue = health-(float)bleed;
			}else{ //regen code
			barvalue = health+(float)regen;
			}
			if(barvalue < 0){barvalue = 0;}if(barvalue > max){barvalue = max;}
			this.statbar.SetValues(barvalue, 0f, max);
		}

		private void ComposeGuis()
		{
			double[] array = new double[]
			{
				1,
				0.0,
				0.0,
				0.5
			};
			IAttribute attribute = this.capi.World.Player.Entity.WatchedAttributes["health"];
			ElementBounds elementBounds = new ElementBounds
			{
				Alignment = EnumDialogArea.CenterBottom,
				BothSizing = ElementSizing.Fixed,
				fixedWidth = 850.0,
				fixedHeight = 100.0
			}.WithFixedAlignmentOffset(0.0, 5.0);
			bool BloodBarHorizontalAlignmentRight = false;
			EnumDialogArea alignment = BloodBarHorizontalAlignmentRight ? EnumDialogArea.RightTop : EnumDialogArea.LeftTop;
			double x = BloodBarHorizontalAlignmentRight ? -2.0 : 1.0;
			ElementBounds elementBounds2 = ElementStdBounds.Statbar(alignment, 348.5).WithFixedAlignmentOffset(x,BloodBarVerticalAlignmentOffset).WithFixedHeight(10.0);
			ElementBounds bounds = elementBounds.FlatCopy().FixedGrow(0.0, 20.0);
			GuiComposer guiComposer = this.capi.Gui.CreateCompo("MD:statbar", bounds);
			this.statbar = new GuiElementStatbar(guiComposer.Api, elementBounds2, array, BloodBarHorizontalAlignmentRight,true);
			guiComposer.BeginChildElements(elementBounds).AddIf(attribute != null).AddInteractiveElement(this.statbar, "Bloodstatsbar").EndIf().EndChildElements().Compose(true);
			this.Composers["MD:BloodBar"] = guiComposer;
			this.TryOpen();
		}

		public override void OnOwnPlayerDataReceived()
		{
			this.ComposeGuis();
			this.UpdateBloodBar();
		}

		public override void OnRenderGUI(float deltaTime)
		{
			if (this.capi.World.Player.WorldData.CurrentGameMode == EnumGameMode.Spectator)
			{
				return;
			}
			base.OnRenderGUI(deltaTime);
		}
		public override EnumDialogType DialogType
		{
			get
			{
				return EnumDialogType.HUD;
			}
		}

		public override bool TryClose()
		{
			return false;
		}

		public override bool ShouldReceiveKeyboardEvents()
		{
			return false;
		}

		public override bool Focusable
		{
			get
			{
				return false;
			}
		}

		private GuiElementStatbar statbar;
	}
}
